from dataclasses import dataclass


@dataclass
class Command:
    name: str
    url: str = "remote"
    api_version: str = "v1"

    def __repr__(self):
        return self.name


COMMAND_ENGINE_ON = Command(name="REON")
COMMAND_ENGINE_OFF = Command(name="REOFF")
COMMAND_COMFORT_ON = Command(name="ROCOMFORTON", api_version="v2")
COMMAND_COMFORT_OFF = Command(name="ROCOMFORTOFF", api_version="v2")
COMMAND_HVAC_ON = Command(name="ROHVACON", api_version="v2")
COMMAND_HVAC_OFF = Command(name="ROHVACOFF", api_version="v2")
COMMAND_PRECOND_ON = Command(name="ROPRECOND")
COMMAND_PRECOND_OFF = Command(name="ROPRECOND_OFF")
COMMAND_LIGHTS_HORN = Command(name="HBLF")
COMMAND_LIGHTS = Command(name="ROLIGHTS")
COMMAND_DOORS_UNLOCK = Command(name="RDU")
COMMAND_DOORS_LOCK = Command(name="RDL")
COMMAND_TRUNK_UNLOCK = Command(name="ROTRUNKUNLOCK", api_version="v2")
COMMAND_TRUNK_LOCK = Command(name="ROTRUNKLOCK", api_version="v2")
COMMAND_LIFTGATE_UNLOCK = Command(name="ROLIFTGATEUNLOCK", api_version="v2")
COMMAND_LIFTGATE_LOCK = Command(name="ROLIFTGATELOCK", api_version="v2")
COMMAND_CABIN_VENTILATION = Command(name="ACV", api_version="v2")
COMMAND_HVAC_TARGET_TEMP = Command(name="ROHVACTMP", api_version="v2")
COMMAND_CHARGE = Command(name="CNOW", url="ev/chargenow")
COMMAND_CHARGE_V4 = Command(name="CNOW2", url="ev/chargenow", api_version="v4")
COMMAND_DEEP_REFRESH = Command(name="DEEPREFRESH", url="ev")
COMMAND_DEEP_REFRESH_V2 = Command(name="DEEPREFRESH2", url="ev", api_version="v2")
COMMAND_REFRESH_LOCATION = Command(name="VF", url="location")

COMMANDS = [
    COMMAND_ENGINE_ON,
    COMMAND_ENGINE_OFF,
    COMMAND_COMFORT_ON,
    COMMAND_COMFORT_OFF,
    COMMAND_HVAC_ON,
    COMMAND_HVAC_OFF,
    COMMAND_PRECOND_ON,
    COMMAND_PRECOND_OFF,
    COMMAND_LIGHTS_HORN,
    COMMAND_LIGHTS,
    COMMAND_DOORS_UNLOCK,
    COMMAND_DOORS_LOCK,
    COMMAND_TRUNK_UNLOCK,
    COMMAND_TRUNK_LOCK,
    COMMAND_LIFTGATE_UNLOCK,
    COMMAND_LIFTGATE_LOCK,
    COMMAND_CABIN_VENTILATION,
    COMMAND_HVAC_TARGET_TEMP,
    COMMAND_CHARGE,
    COMMAND_CHARGE_V4,
    COMMAND_DEEP_REFRESH,
    COMMAND_DEEP_REFRESH_V2,
    COMMAND_REFRESH_LOCATION,
]

COMMANDS_BY_NAME = {x.name: x for x in COMMANDS}



import requests
import uuid
import json
import boto3
import base64
import logging
import http.client as http_client

from dataclasses import dataclass
from dataclasses_json import dataclass_json

from datetime import datetime, timedelta
from requests_auth_aws_sigv4 import AWSSigV4

from .command import Command
from .brands import Brand

_LOGGER = logging.getLogger("py_uconnect")


@dataclass_json
@dataclass
class ChargingLevel:
    name: str


CHARGING_LEVEL_ONE = ChargingLevel("LEVEL_ONE")
CHARGING_LEVEL_TWO = ChargingLevel("LEVEL_TWO")
CHARGING_LEVEL_THREE = ChargingLevel("LEVEL_THREE")
CHARGING_LEVEL_FOUR = ChargingLevel("LEVEL_FOUR")
CHARGING_LEVEL_FIVE = ChargingLevel("LEVEL_FIVE")

CHARGING_LEVELS = [
    CHARGING_LEVEL_ONE,
    CHARGING_LEVEL_TWO,
    CHARGING_LEVEL_THREE,
    CHARGING_LEVEL_FOUR,
    CHARGING_LEVEL_FIVE,
]

CHARGING_LEVELS_BY_NAME = {x.name: x for x in CHARGING_LEVELS}


class API:
    def __init__(
        self,
        email: str,
        password: str,
        pin: str,
        brand: Brand,
        disable_tls_verification: bool = False,
        dev_mode: bool = False,
        trace: bool = False,
    ):
        self.email = email
        self.password = password
        self.pin = pin
        self.brand = brand
        self.dev_mode = dev_mode

        self.uid: str | None = None
        self.aws_auth: AWSSigV4 | None = None

        self.sess = requests.Session()
        self.cognito_client = None

        self.expire_time: datetime | None = None

        if disable_tls_verification:
            self.sess.verify = False
            import urllib3

            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        if trace:
            http_client.HTTPConnection.debuglevel = 1
            logging.basicConfig()
            logging.getLogger().setLevel(logging.DEBUG)
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True

    def _with_default_params(self, params: dict):
        return params | {
            "targetEnv": "jssdk",
            "loginMode": "standard",
            "sdk": "js_latest",
            "authMode": "cookie",
            "sdkBuild": "12234",
            "format": "json",
            "APIKey": self.brand.login_api_key,
        }

    def _default_aws_headers(self, key: str):
        return {
            "x-clientapp-name": "CWP",
            "x-clientapp-version": "1.0",
            "clientrequestid": uuid.uuid4().hex.upper()[0:16],
            "x-api-key": key,
            "locale": self.brand.locale,
            "x-originator-type": "web",
        }

    def set_debug(self, debug: bool):
        _LOGGER.setLevel(logging.DEBUG if debug else logging.WARNING)

    def set_tls_verification(self, verify: bool):
        self.sess.verify = verify

    def set_pin(self, pin: str):
        self.pin = pin

    def login(self):
        """Logs into the Uconnect and caches the auth tokens"""

        if self.cognito_client is None:
            self.cognito_client = boto3.client("cognito-identity", self.brand.region)

        r = self.sess.request(
            method="GET",
            url=self.brand.login_url + "/accounts.webSdkBootstrap",
            params={"apiKey": self.brand.login_api_key},
        )

        r.raise_for_status()
        _LOGGER.debug(f"Login: accounts.webSdkBootstrap: {r.text}")
        r = r.json()

        if r["statusCode"] != 200:
            raise Exception(f"bootstrap failed: {r}")

        r = self.sess.request(
            method="POST",
            url=self.brand.login_url + "/accounts.login",
            params=self._with_default_params(
                {
                    "loginID": self.email,
                    "password": self.password,
                    "sessionExpiration": 300,
                    "include": "profile,data,emails,subscriptions,preferences",
                }
            ),
        )

        r.raise_for_status()
        _LOGGER.debug(f"Login: accounts.login: {r.text}")
        r = r.json()

        if r["statusCode"] != 200:
            raise Exception(f"account login failed: {r}")

        self.uid = r["UID"]
        login_token = r["sessionInfo"]["login_token"]

        r = self.sess.request(
            method="POST",
            url=self.brand.login_url + "/accounts.getJWT",
            params=self._with_default_params(
                {
                    "login_token": login_token,
                    "fields": "profile.firstName,profile.lastName,profile.email,country,locale,data.disclaimerCodeGSDP",
                }
            ),
        )

        r.raise_for_status()
        _LOGGER.debug(f"Login: accounts.getJWT: {r.text}")
        r = r.json()

        if r["statusCode"] != 200:
            raise Exception(f"unable to obtain JWT: {r}")

        r = self.sess.request(
            method="POST",
            url=self.brand.token_url,
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            json={"gigya_token": r["id_token"]},
        )

        r.raise_for_status()
        _LOGGER.debug(f"Login: obtain token: {r.text}")
        r = r.json()

        token = r.get("Token", None)
        identity_id = r.get("IdentityId", None)
        if token is None or identity_id is None:
            raise Exception(f"unable to obtain identity & token: {r}")

        r = self.cognito_client.get_credentials_for_identity(
            IdentityId=identity_id,
            Logins={"cognito-identity.amazonaws.com": token},
        )

        creds = r.get("Credentials", None)
        if not creds:
            raise Exception(f"unable to obtain AWS credentials: {r}")

        self.aws_auth = AWSSigV4(
            "execute-api",
            region=self.brand.region,
            aws_access_key_id=creds["AccessKeyId"],
            aws_secret_access_key=creds["SecretKey"],
            aws_session_token=creds["SessionToken"],
        )

        self.expire_time = creds["Expiration"]

    def _refresh_token_if_needed(self):
        """Checks if token is available and fresh, refreshes it otherwise"""

        if self.dev_mode:
            return

        if (
            self.expire_time is None
            or datetime.now().astimezone() > self.expire_time - timedelta(minutes=5)
        ):
            try:
                self.login()
            except Exception as e:
                raise Exception(f"unable to login: {e}")

    def list_vehicles(self) -> list[dict]:
        """Loads a list of vehicles with general info"""

        if self.dev_mode:
            with open("test_list.json") as f:
                return json.load(f)["vehicles"]

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url + f"/v4/accounts/{self.uid}/vehicles",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            params={"stage": "ALL", "sdp": "ALL", "brand": self.brand.brand_code},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"list_vehicles: {r.text}")
        r = r.json()

        if "vehicles" not in r:
            raise Exception(f"incorrect response: {r}")

        return r["vehicles"]

    def get_vehicle(self, vin: str) -> dict:
        """Gets a more detailed info abount a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_vehicle_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url + f"/v3/accounts/{self.uid}/vehicles/{vin}/status",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle ({vin}): {r.text}")
        r = r.json()

        return r

    def get_vehicle_status(self, vin: str) -> dict:
        """Loads another part of status of a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_vehicle_status_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/remote/status",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle_status ({vin}): {r.text}")
        r = r.json()

        return r

    def get_vehicle_location(self, vin: str) -> dict:
        """Gets last known location of a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_vehicle_location_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/location/lastknown",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle_location ({vin}): {r.text}")
        r = r.json()

        return r

    def get_vehicle_health_report(self, vin: str) -> dict:
        """Gets vehicle health report for a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_vehicle_vhr_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url + f"/v1/accounts/{self.uid}/vehicles/{vin}/vhr/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle_health_report ({vin}): {r.text}")
        r = r.json()

        return r

    def get_maintenance_history(self, vin: str) -> dict:
        """Gets maintenance history for a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_maintenance_history_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/maintenance/history/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_maintenance_history ({vin}): {r.text}")
        r = r.json()

        return r

    def get_eco_coaching_last_trip(self, vin: str) -> dict:
        """Gets eco-coaching data for the last trip of a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_eco_coaching_last_trip_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url
            + f"/v2/accounts/{self.uid}/vehicles/{vin}/ecocoaching/get-last-trip/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_eco_coaching_last_trip ({vin}): {r.text}")
        r = r.json()

        return r

    def get_eco_coaching_trips(self, vin: str) -> dict:
        """Gets eco-coaching trip list for a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_eco_coaching_trips_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url
            + f"/v2/accounts/{self.uid}/vehicles/{vin}/ecocoaching/get-trips/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_eco_coaching_trips ({vin}): {r.text}")
        r = r.json()

        return r

    def get_vehicle_image(self, vin: str, width: int = 600, height: int = 340) -> dict:
        """Gets vehicle image URL for a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_vehicle_image_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        data = {
            "imageURLs": [{"id": vin, "width": width, "height": height, "resp": "png"}]
        }

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url + f"/v4/accounts/{self.uid}/vehicles/{vin}/image/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json=data,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle_image ({vin}): {r.text}")
        r = r.json()

        return r

    def get_vehicle_notifications(self, vin: str, limit: int | None = 30) -> dict:
        """Loads notifications for a vehicle with a given VIN"""

        self._refresh_token_if_needed()

        url = (
            self.brand.api.url + f"/v1/accounts/{self.uid}/vehicles/{vin}/notifications"
        )

        if limit is not None:
            url += f"?limit={limit}"

        r = self.sess.request(
            method="GET",
            url=url,
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle_notifications ({vin}): {r.text}")
        r = r.json()

        return r

    def get_stolen_vehicle_status(self, vin: str) -> dict:
        """Gets stolen vehicle locator (SVLA) status for a vehicle with a given VIN"""

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/svla/status/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_stolen_vehicle_status ({vin}): {r.text}")
        r = r.json()

        return r

    def get_vehicle_subscription(self, vin: str) -> dict:
        """Gets subscription status for a vehicle with a given VIN"""

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/subscription/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_vehicle_subscription ({vin}): {r.text}")
        r = r.json()

        return r

    def set_vehicle_nickname(self, vin: str, nickname: str) -> dict:
        """Sets a nickname for a vehicle with a given VIN"""

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/nickname/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json={"nickname": nickname},
        )

        r.raise_for_status()
        _LOGGER.debug(f"set_vehicle_nickname ({vin} {nickname}): {r.text}")
        r = r.json()

        return r

    def update_location(self, vin: str) -> str:
        """Triggers a fresh location update for a vehicle with a given VIN.

        Returns the correlation ID to poll for completion.
        Use get_vehicle_location() afterwards to retrieve the updated location.
        """

        pin_auth = self._pin_auth()

        data = {
            "command": "VF",
            "pinAuth": pin_auth,
        }

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/location/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json=data,
        )

        r.raise_for_status()
        _LOGGER.debug(f"update_location ({vin}): {r.text}")
        r = r.json()

        if "responseStatus" not in r or r["responseStatus"] != "pending":
            error = r.get("debugMsg", "unknown error")
            raise Exception(f"update location failed: {error} ({r})")

        return r["correlationId"]

    def get_remote_operation_status(self, vin: str, correlation_id: str) -> dict:
        """Gets the status of a remote operation by its correlation ID"""

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v1/accounts/{self.uid}/vehicles/{vin}/remote/{correlation_id}/status/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_remote_operation_status ({vin} {correlation_id}): {r.text}")
        r = r.json()

        return r

    def _pin_auth(self) -> str:
        data = {
            "pin": base64.b64encode(self.pin.encode()).decode(encoding="utf-8"),
        }

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="POST",
            url=self.brand.auth.url
            + f"/v1/accounts/{self.uid}/ignite/pin/authenticate",
            headers=self._default_aws_headers(self.brand.auth.token)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json=data,
        )

        r.raise_for_status()
        _LOGGER.debug(f"pin auth: {r.text}")
        r = r.json()

        if "token" not in r:
            raise Exception(f"authentication failed: no token found: {r}")

        return r["token"]

    def command(self, vin: str, cmd: Command):
        """Sends given command to the vehicle with a given VIN"""

        if self.dev_mode:
            return

        pin_auth = self._pin_auth()

        data = {
            "command": cmd.name,
            "pinAuth": pin_auth,
        }

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url
            + f"/{cmd.api_version}/accounts/{self.uid}/vehicles/{vin}/{cmd.url}",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json=data,
        )

        r.raise_for_status()
        _LOGGER.debug(f"command execute ({vin} {cmd}): {r.text}")
        r = r.json()

        if "responseStatus" not in r or r["responseStatus"] != "pending":
            error = r.get("debugMsg", "unknown error")
            raise Exception(f"command queuing failed: {error} ({r})")

        return r["correlationId"]

    def get_charge_schedules(self, vin: str) -> dict:
        """Gets EV charge schedules for a vehicle with a given VIN"""

        if self.dev_mode:
            with open(f"test_charge_schedules_{vin}.json") as f:
                return json.load(f)

        self._refresh_token_if_needed()

        r = self.sess.request(
            method="GET",
            url=self.brand.api.url
            + f"/v4/accounts/{self.uid}/vehicles/{vin}/ev/schedule/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
        )

        r.raise_for_status()
        _LOGGER.debug(f"get_charge_schedules ({vin}): {r.text}")
        r = r.json()

        return r

    def set_charge_schedule(self, vin: str, schedule: dict):
        """Sets an EV charge schedule on the vehicle with a given VIN"""

        if self.dev_mode:
            return

        pin_auth = self._pin_auth()

        data = schedule | {"pinAuth": pin_auth}

        r = self.sess.request(
            method="POST",
            url=self.brand.api.url
            + f"/v4/accounts/{self.uid}/vehicles/{vin}/ev/schedule/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json=data,
        )

        r.raise_for_status()
        _LOGGER.debug(f"set_charge_schedule ({vin}): {r.text}")
        r = r.json()

        if "correlationId" not in r:
            error = r.get("debugMsg", "unknown error")
            raise Exception(f"set charge schedule failed: {error} ({r})")

        return r["correlationId"]

    def set_charging_level(
        self, vin: str, level: ChargingLevel, max_soc: str | None = None
    ):
        """Sets the charging level on the vehicle with a given VIN"""

        if self.dev_mode:
            return

        pin_auth = self._pin_auth()

        data = {
            "preference": level.name,
            "pinAuth": pin_auth,
        }

        if max_soc is not None:
            data["maxSOC"] = max_soc

        r = self.sess.request(
            method="PUT",
            url=self.brand.api.url
            + f"/v2/accounts/{self.uid}/vehicles/{vin}/ev/charge/preference/",
            headers=self._default_aws_headers(self.brand.api.key)
            | {"content-type": "application/json"},
            auth=self.aws_auth,
            json=data,
        )

        r.raise_for_status()
        _LOGGER.debug(f"set charging level ({vin} {level.name}): {r.text}")
        r = r.json()

        if "correlationId" not in r:
            error = r.get("debugMsg", "unknown error")
            raise Exception(f"set charging level failed: {error} ({r})")

        return r["correlationId"]



from dataclasses import dataclass

API_URL_EU = "https://channels.sdpr-01.fcagcv.com"
API_URL_US = "https://channels.sdpr-02.fcagcv.com"

TOKEN_URL_EU = "https://authz.sdpr-01.fcagcv.com/v2/cognito/identity/token"
TOKEN_URL_US = "https://authz.sdpr-02.fcagcv.com/v2/cognito/identity/token"
TOKEN_URL_US_PREP = "https://authz.sdpr-02.prep.fcagcv.com/v2/cognito/identity/token"

LOCALE_EU = "de_de"
LOCALE_US = "en_us"

REGION_US = "us-east-1"
REGION_EU = "eu-west-1"


@dataclass
class API:
    url: str
    key: str


API_EU = API(
    url="https://channels.sdpr-01.fcagcv.com",
    key="2wGyL6PHec9o1UeLPYpoYa1SkEWqeBur9bLsi24i",
)

API_US = API(
    url="https://channels.sdpr-02.fcagcv.com",
    key="OgNqp2eAv84oZvMrXPIzP8mR8a6d9bVm1aaH9LqU",
)


@dataclass
class Auth:
    url: str
    token: str


AUTH_EU = Auth(
    url="https://mfa.fcl-01.fcagcv.com",
    token="JWRYW7IYhW9v0RqDghQSx4UcRYRILNmc8zAuh5ys",
)

AUTH_US = Auth(
    url="https://mfa.fcl-02.fcagcv.com",
    token="fNQO6NjR1N6W0E5A6sTzR3YY4JGbuPv48Nj9aZci",
)

AUTH_US_PREP = Auth(
    url="https://mfa.fcl-02.prep.fcagcv.com",
    token="lHBEtsqT1Y5oKvzhvA9KW6rkirU3ZtGf44jTIiQV",
)


@dataclass
class Brand:
    name: str
    region: str
    login_api_key: str
    login_url: str
    token_url: str
    api: API
    auth: Auth
    locale: str
    brand_code: str

    def __repr__(self):
        return self.name


FIAT_EU = Brand(
    name="FIAT_EU",
    region=REGION_EU,
    login_api_key="3_mOx_J2dRgjXYCdyhchv3b5lhi54eBcdCTX4BI8MORqmZCoQWhA0mV2PTlptLGUQI",
    login_url="https://loginmyuconnect.fiat.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="FIAT",
)

FIAT_US = Brand(
    name="FIAT_US",
    region=REGION_US,
    login_api_key="3_WfFvlZJwcSdOD0LFQCngUV3W390R4Yshpuq3RsZvnV4VG0c9Q6R0RtDwcXc8dTrI",
    login_url="https://login-us.fiat.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

FIAT_ASIA = Brand(
    name="FIAT_ASIA",
    region=REGION_EU,
    login_api_key="4_YAQNaPqdPEUbbzhvhunKAA",
    login_url="https://login-iap.fiat.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="FIAT",
)

FIAT_CANADA = Brand(
    name="FIAT_CANADA",
    region=REGION_US,
    login_api_key="3_Ii2kSgQm4ljy19LIZeLwa76OlmWbpSa8w3aSP5VJdx19tub3oWxsFR-HEusDnUEh",
    login_url="https://login-stage-us.fiat.com",
    token_url=TOKEN_URL_US_PREP,
    api=API_US,
    auth=AUTH_US_PREP,
    locale=LOCALE_US,
    brand_code="ALL",
)

ALFA_ROMEO_US_CANADA = Brand(
    name="ALFA_ROMEO_US_CANADA",
    region=REGION_US,
    login_api_key="3_FSxGyaktviayTDRcgp9r9o2KjuFSrHT13wWNN9zPrvAGUCoXPDqoIPOwlBUhck4A",
    login_url="https://login-us.alfaromeo.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

ALFA_ROMEO_ASIA = Brand(
    name="ALFA_ROMEO_ASIA",
    region=REGION_EU,
    login_api_key="4_PSQeADnQ4p5XOaDgT0B5pA",
    login_url="https://login-iap.alfaromeo.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="ALFAROMEO",
)

ALFA_ROMEO_EU = Brand(
    name="ALFA_ROMEO_EU",
    region=REGION_EU,
    login_api_key="3_h8sj2VQI-KYXiunPq9a1QuAA4yWkY0r5AD1u8A8B1RPn_Cvl54xcoc2-InH5onJ1",
    login_url="https://login.alfaromeo.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="ALFAROMEO",
)

CHRYSLER_CANADA = Brand(
    name="CHRYSLER_CANADA",
    region=REGION_US,
    login_api_key="3_gdhu-ur4jc2hEryDMnF4YPELkjzSi-invZTjop4isZu4ReHodVcuL44u93cOUqMC",
    login_url="https://login-stage-us.chrysler.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

CHRYSLER_US = Brand(
    name="CHRYSLER_US",
    region=REGION_US,
    login_api_key="3_cv4AzHkJh48-cqwaf_Ahcg1HnsmQqz1lm0sOdVdHW5FjT3m6SyywywOBaskBQqwn",
    login_url="https://login-us.chrysler.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

MASERATI_EU = Brand(
    name="MASERATI_EU",
    region=REGION_EU,
    login_api_key="3_rNbVuhn2gIt3BnLjlGsJcMo26Lft3avDne_FLRT34Dy_9OxHtCVOnplwY436lGZa",
    login_url="https://login.maserati.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="MASERATI",
)

MASERATI_ASIA = Brand(
    name="MASERATI_ASIA",
    region=REGION_EU,
    login_api_key="4_uwF-in6KF-aMbEkPAb-fOg",
    login_url="https://accounts.au1.gigya.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="MASERATI",
)

MASERATI_US_CANADA = Brand(
    name="MASERATI_US_CANADA",
    region=REGION_US,
    login_api_key="3_nShL4-O7IL0OGqroO8AzwiRU0-ZHcBZ4TLBrh5MORusMo5XYxhCLXPYfjI4OOLOy",
    login_url="https://login-us.maserati.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

JEEP_EU = Brand(
    name="JEEP_EU",
    region=REGION_EU,
    login_api_key="3_ZvJpoiZQ4jT5ACwouBG5D1seGEntHGhlL0JYlZNtj95yERzqpH4fFyIewVMmmK7j",
    login_url="https://login.jeep.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="REST",
)

JEEP_US = Brand(
    name="JEEP_US",
    region=REGION_US,
    login_api_key="3_5qxvrevRPG7--nEXe6huWdVvF5kV7bmmJcyLdaTJ8A45XUYpaR398QNeHkd7EB1X",
    login_url="https://login-us.jeep.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

JEEP_ASIA = Brand(
    name="JEEP_ASIA",
    region=REGION_EU,
    login_api_key="4_zqGYHC7rM8RCHHl4YFDebA",
    login_url="https://login-iap.jeep.com",
    token_url=TOKEN_URL_EU,
    api=API_EU,
    auth=AUTH_EU,
    locale=LOCALE_EU,
    brand_code="REST",
)

DODGE_US = Brand(
    name="DODGE_US",
    region=REGION_US,
    login_api_key="4_dSRvo6ZIpp8_St7BF9VHGA",
    login_url="https://login-us.dodge.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

RAM_US = Brand(
    name="RAM_US",
    region=REGION_US,
    login_api_key="3_7YjzjoSb7dYtCP5-D6FhPsCciggJFvM14hNPvXN9OsIiV1ujDqa4fNltDJYnHawO",
    login_url="https://login-us.ramtrucks.com",
    token_url=TOKEN_URL_US,
    api=API_US,
    auth=AUTH_US,
    locale=LOCALE_US,
    brand_code="ALL",
)

BRANDS = {
    FIAT_EU.name: FIAT_EU,
    FIAT_US.name: FIAT_US,
    FIAT_CANADA.name: FIAT_CANADA,
    FIAT_ASIA.name: FIAT_ASIA,
    JEEP_EU.name: JEEP_EU,
    JEEP_US.name: JEEP_US,
    JEEP_ASIA.name: JEEP_ASIA,
    DODGE_US.name: DODGE_US,
    RAM_US.name: RAM_US,
    CHRYSLER_CANADA.name: CHRYSLER_CANADA,
    CHRYSLER_US.name: CHRYSLER_US,
    ALFA_ROMEO_US_CANADA.name: ALFA_ROMEO_US_CANADA,
    ALFA_ROMEO_EU.name: ALFA_ROMEO_EU,
    ALFA_ROMEO_ASIA.name: ALFA_ROMEO_ASIA,
    MASERATI_ASIA.name: MASERATI_ASIA,
    MASERATI_EU.name: MASERATI_EU,
    MASERATI_US_CANADA.name: MASERATI_US_CANADA,
}



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    author="Olga Novgorodova",
    author_email="olga@novg.net",
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
    ],
    description="Python API Client for FCA/Stellantis cars",
    install_requires=[
        "boto3>=1.35.96",
        "requests>=2.32.3",
        "requests-auth-aws-sigv4>=0.7",
        "dataclasses_json>=0.6.7",
    ],
    license="Apache license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords=[
        "fiat",
        "jeep",
        "dodge",
        "ram",
        "alfa",
        "romeo",
        "maserati",
        "chrysler",
        "api",
        "cloud",
    ],
    name="py-uconnect",
    packages=find_packages(include=["py_uconnect", "py_uconnect.*"]),
    url="https://github.com/hass-uconnect/py-uconnect",
    version="0.4.0",
    zip_safe=False,
)



from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from typing import Any, Dict
from datetime import datetime, timedelta
from time import sleep

from .api import API, ChargingLevel
from .brands import Brand
from .command import Command, COMMANDS_BY_NAME


def convert(v: Any) -> Any:
    if not isinstance(v, str):
        return v

    if v == "null":
        return None

    try:
        v = int(v)
    except (ValueError, TypeError):
        try:
            v = float(v)
        except (ValueError, TypeError):
            pass

    return v


def sg(dct: Any, *keys: str) -> Any:
    if not isinstance(dct, dict):
        return None

    for key in keys:
        try:
            dct = dct[key]
        except KeyError:
            return None

    return convert(dct)


def sg_eq(dct: Any, expect: Any, *keys: str) -> bool | None:
    v = sg(dct, *keys)

    if v is None:
        return None

    return v == expect


def sg_eq_str(dct: Any, expect: str, *keys: str) -> bool:
    v = sg(dct, *keys)

    if not isinstance(v, str):
        return False

    return v.lower() == expect.lower()


CHARGING_LEVELS = {
    "DEFAULT": 0,
    "LEVEL_1": 1,
    "LEVEL_2": 2,
    "LEVEL_3": 3,
}


@dataclass_json
@dataclass
class Location:
    longitude: float | None = None
    latitude: float | None = None
    altitude: float | None = None
    bearing: float | None = None
    is_approximate: bool | None = None
    updated: datetime | None = None

    def __repr__(self):
        return f"lat: {self.latitude}, lon: {self.longitude} (updated {self.updated})"


@dataclass_json
@dataclass
class Vehicle:
    vin: str

    # General info
    nickname: str
    make: str
    model: str
    year: str
    region: str

    sdp: str | None = None
    image_url: str | None = None
    fuel_type: str | None = None

    # Status
    ignition_on: bool | None = None
    trunk_locked: bool | None = None

    odometer: float | None = None
    odometer_unit: str | None = None
    days_to_service: int | None = None
    distance_to_service: int | None = None
    distance_to_service_unit: str | None = None
    distance_to_empty: int | None = None
    distance_to_empty_unit: str | None = None
    range_gas: int | None = None
    range_gas_unit: str | None = None
    range_total: int | None = None
    range_total_unit: str | None = None
    battery_voltage: float | None = None
    battery_state_of_charge: str | None = None
    oil_level: int | None = None
    fuel_low: bool | None = None
    fuel_amount: int | None = None

    # EV related
    plugged_in: bool | None = None
    ev_running: bool | None = None
    charging: bool | None = None
    charging_level: int | None = None
    charging_level_preference: str | None = None
    state_of_charge: float | None = None
    time_to_fully_charge_l3: int | None = None
    time_to_fully_charge_l2: int | None = None
    time_to_fully_charge_l1: int | None = None
    ev_head_seat: bool | None = None
    ev_cabin_cond: bool | None = None

    # Wheels
    wheel_front_left_pressure: float | None = None
    wheel_front_left_pressure_unit: str | None = None
    wheel_front_left_pressure_warning: bool | None = None
    wheel_front_right_pressure: float | None = None
    wheel_front_right_pressure_unit: str | None = None
    wheel_front_right_pressure_warning: bool | None = None
    wheel_rear_left_pressure: float | None = None
    wheel_rear_left_pressure_unit: str | None = None
    wheel_rear_left_pressure_warning: bool | None = None
    wheel_rear_right_pressure: float | None = None
    wheel_rear_right_pressure_unit: str | None = None
    wheel_rear_right_pressure_warning: bool | None = None

    # Doors
    door_driver_locked: bool | None = None
    door_passenger_locked: bool | None = None
    door_rear_left_locked: bool | None = None
    door_rear_right_locked: bool | None = None

    # Windows
    window_driver_closed: bool | None = None
    window_passenger_closed: bool | None = None

    location: Location | None = None
    supported_commands: list[str] = field(default_factory=list)
    enabled_services: list[str] = field(default_factory=list)

    timestamp_info: datetime | None = None
    timestamp_status: datetime | None = None

    def __repr__(self):
        return f"{self.vin} (nick: {self.nickname})"


def _update_vehicle(v: Vehicle, p: dict) -> Vehicle:
    vi = sg(p, "vehicleInfo")
    ev = sg(p, "evInfo")
    batt = sg(ev, "battery")

    v.battery_voltage = sg(vi, "batteryInfo", "batteryVoltage", "value")
    v.battery_state_of_charge = sg(vi, "batteryInfo", "batteryStateOfCharge")
    v.charging = sg_eq(batt, "CHARGING", "chargingStatus")
    v.charging_level = CHARGING_LEVELS.get(sg(batt, "chargingLevel"), None)
    v.charging_level_preference = sg(ev, "chargePowerPreference")
    if v.charging_level_preference == "NOT_USED_VALUE":
        v.charging_level_preference = None

    v.plugged_in = sg(batt, "plugInStatus")
    v.state_of_charge = sg(batt, "stateOfCharge")

    v.days_to_service = sg(vi, "daysToService")
    v.distance_to_service = sg(vi, "distanceToService", "distanceToService", "value")
    v.distance_to_service_unit = sg(
        vi, "distanceToService", "distanceToService", "unit"
    )
    v.distance_to_empty = sg(batt, "distanceToEmpty", "value") or sg(
        vi, "fuel", "distanceToEmpty", "value"
    )
    v.distance_to_empty_unit = sg(batt, "distanceToEmpty", "unit") or sg(
        vi, "fuel", "distanceToEmpty", "unit"
    )
    v.range_gas = sg(batt, "gasRange")
    v.range_gas_unit = v.distance_to_empty_unit
    v.range_total = sg(batt, "totalRange")
    v.range_total_unit = v.distance_to_empty_unit

    v.fuel_low = sg(vi, "fuel", "isFuelLevelLow")
    v.fuel_amount = sg(vi, "fuel", "fuelAmountLevel")
    v.oil_level = sg(vi, "oilLevel", "oilLevel")

    v.ignition_on = sg_eq(ev, "ON", "ignitionStatus")

    csai = sg(ev, "chargeScheduleAdditionalInfo")
    v.ev_head_seat = sg(csai, "headSeat")
    v.ev_cabin_cond = sg(csai, "cabinCond")

    v.time_to_fully_charge_l3 = sg(batt, "timeToFullyChargeL3")
    v.time_to_fully_charge_l2 = sg(batt, "timeToFullyChargeL2")
    v.time_to_fully_charge_l1 = sg(batt, "timeToFullyChargeL1")

    # Some vehicles report -1
    if v.time_to_fully_charge_l3 is not None and v.time_to_fully_charge_l3 < 0:
        v.time_to_fully_charge_l3 = None
    if v.time_to_fully_charge_l2 is not None and v.time_to_fully_charge_l2 < 0:
        v.time_to_fully_charge_l2 = None
    if v.time_to_fully_charge_l1 is not None and v.time_to_fully_charge_l1 < 0:
        v.time_to_fully_charge_l1 = None

    v.odometer = sg(vi, "odometer", "odometer", "value")
    v.odometer_unit = sg(vi, "odometer", "odometer", "unit")

    if isinstance(vi, dict) and "tyrePressure" in vi:
        tp = {x["type"]: x for x in vi["tyrePressure"]}

        v.wheel_front_left_pressure = sg(tp, "FL", "pressure", "value")
        v.wheel_front_left_pressure_unit = sg(tp, "FL", "pressure", "unit")
        v.wheel_front_left_pressure_warning = sg(tp, "FL", "warning")

        v.wheel_front_right_pressure = sg(tp, "FR", "pressure", "value")
        v.wheel_front_right_pressure_unit = sg(tp, "FR", "pressure", "unit")
        v.wheel_front_right_pressure_warning = sg(tp, "FR", "warning")

        v.wheel_rear_left_pressure = sg(tp, "RL", "pressure", "value")
        v.wheel_rear_left_pressure_unit = sg(tp, "RL", "pressure", "unit")
        v.wheel_rear_left_pressure_warning = sg(tp, "RL", "warning")

        v.wheel_rear_right_pressure = sg(tp, "RR", "pressure", "value")
        v.wheel_rear_right_pressure_unit = sg(tp, "RR", "pressure", "unit")
        v.wheel_rear_right_pressure_warning = sg(tp, "RR", "warning")

    v.timestamp_info = (
        datetime.fromtimestamp(p["timestamp"] / 1000).astimezone()
        if "timestamp" in p
        else None
    )

    return v


class Client:
    def __init__(
        self,
        email: str,
        password: str,
        pin: str,
        brand: Brand,
        disable_tls_verification: bool = False,
        dev_mode: bool = False,
        trace: bool = False,
    ):
        self.api = API(
            email,
            password,
            pin,
            brand,
            disable_tls_verification=disable_tls_verification,
            dev_mode=dev_mode,
            trace=trace,
        )
        self.vehicles: Dict[str, Vehicle] = {}

    def set_debug(self, debug: bool):
        """Sets debug logging on/off"""

        self.api.set_debug(debug)

    def set_tls_verification(self, verify: bool):
        """Enable or disable TLS certificate verification"""

        self.api.set_tls_verification(verify)

    def set_pin(self, pin: str):
        self.api.set_pin(pin)

    def refresh(self):
        """Refreshes all the vehicle data and caches it locally"""

        vehicles = self.api.list_vehicles()

        for x in vehicles:
            vin = x["vin"]

            if vin not in self.vehicles:
                vehicle = Vehicle(
                    vin=vin,
                    nickname=sg(x, "nickname"),
                    make=sg(x, "make"),
                    model=sg(x, "modelDescription"),
                    year=sg(x, "tsoModelYear"),
                    region=sg(x, "soldRegion"),
                )
                self.vehicles[vin] = vehicle
            else:
                vehicle = self.vehicles[vin]

            vehicle.sdp = sg(x, "sdp")
            vehicle.image_url = sg(x, "vehicleImageURL")
            vehicle.fuel_type = sg(x, "fuelType")

            info = self.api.get_vehicle(vin)
            _update_vehicle(vehicle, info)

            try:
                loc = self.api.get_vehicle_location(vin)

                updated = (
                    datetime.fromtimestamp(loc["timeStamp"] / 1000).astimezone()
                    if "timeStamp" in loc
                    else None
                )

                vehicle.location = Location(
                    longitude=sg(loc, "longitude"),
                    latitude=sg(loc, "latitude"),
                    altitude=sg(loc, "altitude"),
                    bearing=sg(loc, "bearing"),
                    is_approximate=sg(loc, "isLocationApprox"),
                    updated=updated,
                )
            except Exception:
                pass

            try:
                s = self.api.get_vehicle_status(vin)

                if "doors" in s:
                    vehicle.door_driver_locked = sg_eq(
                        s, "LOCKED", "doors", "driver", "status"
                    )
                    vehicle.door_passenger_locked = sg_eq(
                        s, "LOCKED", "doors", "passenger", "status"
                    )
                    vehicle.door_rear_left_locked = sg_eq(
                        s, "LOCKED", "doors", "leftRear", "status"
                    )
                    vehicle.door_rear_right_locked = sg_eq(
                        s, "LOCKED", "doors", "rightRear", "status"
                    )

                if "windows" in s:
                    vehicle.window_driver_closed = sg_eq(
                        s, "CLOSED", "windows", "driver", "status"
                    )
                    vehicle.window_passenger_closed = sg_eq(
                        s, "CLOSED", "windows", "passenger", "status"
                    )

                if "engine" in s:
                    vehicle.ignition_on = sg_eq(s, "ON", "engine", "status")

                vehicle.trunk_locked = sg_eq(s, "LOCKED", "trunk", "status")
                vehicle.ev_running = sg_eq(s, "ON", "evRunning", "status")

                vehicle.timestamp_status = (
                    datetime.fromtimestamp(s["timestamp"] / 1000).astimezone()
                    if "timestamp" in s
                    else None
                )
            except Exception:
                pass

            enabled_services = []
            if "services" in x:
                enabled_services = [
                    v["service"]
                    for v in x["services"]
                    if sg(v, "vehicleCapable") and sg(v, "serviceEnabled")
                ]

            vehicle.enabled_services = enabled_services
            vehicle.supported_commands = [
                v for v in enabled_services if v in COMMANDS_BY_NAME
            ]

    def get_vehicles(self):
        """Returns all vehicles data. Must execute refresh method before."""

        return self.vehicles

    def _get_commands_statuses(self, vin: str) -> dict:
        r = self.api.get_vehicle_notifications(vin)

        if "notifications" not in r or "items" not in r["notifications"]:
            return {}

        return {
            x["correlationId"]: sg_eq_str(
                x, "success", "notification", "data", "status"
            )
            for x in r["notifications"]["items"]
            if "correlationId" in x
        }

    def _poll_correlation_id(
        self,
        vin: str,
        id: str,
        timeout: timedelta = timedelta(seconds=60),
        interval: timedelta = timedelta(seconds=2),
    ) -> bool:
        start = datetime.now()
        while datetime.now() - start < timeout:
            sleep(interval.seconds)
            r = self._get_commands_statuses(vin)
            if id in r:
                return r[id]

        raise Exception(f"unable to obtain execution status: timed out (id {id})")

    def get_remote_operation_status(self, vin: str, correlation_id: str) -> dict:
        """Get the status of a remote operation by its correlation ID"""

        return self.api.get_remote_operation_status(vin, correlation_id)

    def command(self, vin: str, cmd: Command):
        """Execute a given command against a car with a given VIN"""

        return self.api.command(vin, cmd)

    def command_verify(
        self,
        vin: str,
        cmd: Command,
    ) -> bool:
        """Execute a given command against a car with a given VIN and poll for the status"""

        id = self.command(vin, cmd)
        return self._poll_correlation_id(vin, id)

    def get_stolen_vehicle_status(self, vin: str) -> dict:
        """Get stolen vehicle locator (SVLA) status for a vehicle with a given VIN"""

        return self.api.get_stolen_vehicle_status(vin)

    def get_vehicle_subscription(self, vin: str) -> dict:
        """Get subscription status for a vehicle with a given VIN"""

        return self.api.get_vehicle_subscription(vin)

    def set_vehicle_nickname(self, vin: str, nickname: str) -> dict:
        """Set a nickname for a vehicle with a given VIN"""

        return self.api.set_vehicle_nickname(vin, nickname)

    def update_location(self, vin: str) -> str:
        """Trigger a fresh location update for a vehicle with a given VIN.

        Returns the correlation ID to poll for completion.
        """

        return self.api.update_location(vin)

    def get_eco_coaching_last_trip(self, vin: str) -> dict:
        """Get eco-coaching data for the last trip of a vehicle with a given VIN"""

        return self.api.get_eco_coaching_last_trip(vin)

    def get_eco_coaching_trips(self, vin: str) -> dict:
        """Get eco-coaching trip list for a vehicle with a given VIN"""

        return self.api.get_eco_coaching_trips(vin)

    def get_maintenance_history(self, vin: str) -> dict:
        """Get maintenance history for a vehicle with a given VIN"""

        return self.api.get_maintenance_history(vin)

    def get_vehicle_health_report(self, vin: str) -> dict:
        """Get vehicle health report for a vehicle with a given VIN"""

        return self.api.get_vehicle_health_report(vin)

    def get_vehicle_image(self, vin: str) -> dict:
        """Get vehicle image URL for a vehicle with a given VIN"""

        return self.api.get_vehicle_image(vin)

    def get_charge_schedules(self, vin: str) -> dict:
        """Get EV charge schedules for a vehicle with a given VIN"""

        return self.api.get_charge_schedules(vin)

    def set_charge_schedule(self, vin: str, schedule: dict):
        """Set an EV charge schedule on the vehicle with a given VIN"""

        return self.api.set_charge_schedule(vin, schedule)

    def set_charge_schedule_verify(self, vin: str, schedule: dict) -> bool:
        """Set an EV charge schedule and poll for the status"""

        id = self.api.set_charge_schedule(vin, schedule)
        return self._poll_correlation_id(vin, id)

    def set_charging_level(
        self, vin: str, level: ChargingLevel, max_soc: str | None = None
    ):
        """Set the charging level on the vehicle with a given VIN"""

        return self.api.set_charging_level(vin, level, max_soc=max_soc)

    def set_charging_level_verify(
        self, vin: str, level: ChargingLevel, max_soc: str | None = None
    ):
        """Set the charging level on the vehicle with a given VIN and poll for the status"""

        id = self.api.set_charging_level(vin, level, max_soc=max_soc)
        return self._poll_correlation_id(vin, id)
